// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// Module: PosWCS
//
// Author: Hassan Siddiqui
//
// Description: 
//
// Reads in X/Y WCS from a given dataset.
//
// If the info in invalid/not present, 
// the valid() method will return false.
//
// Modifications: See ChangeLog
// 
// $Id: PosWCS.cc,v 1.1 2011/01/11 16:36:10 rsaxton Exp $

/* General */

/* SAS-specific */
#include "errstr.h"
#include "Tracer.h"

/* Local classes */
#include "PosWCS.h"

/*** Constructors */

/* Read state from argument list */
PosWCS::PosWCS( const DataSet * const set,
		double refra,
		double refdec, double refpa, bool useodf):
//  _valid(true),
  _refra(refra),
  _refdec(refdec),
  _refpa(refpa),
  _useodf(useodf)
{
  PROC;
  /* Get dataset attributes */
  if ( !set ||  
       !set->hasAttribute("REFXCTYP") ||
       !set->hasAttribute("REFXCRPX") ||
       !set->hasAttribute("REFXCRVL") ||
       !set->hasAttribute("REFXCDLT") ||
       !set->hasAttribute("REFYCTYP") ||
       !set->hasAttribute("REFYCRPX") ||
       !set->hasAttribute("REFYCRVL") ||
       !set->hasAttribute("REFYCDLT") ) {
	string ratyp   = "RA---TAN";
	string dectyp  = "DEC--TAN";
	unsigned refpx = 25921;
	double decdlt  = 0.05/3600; 
	double radlt  = -1 * decdlt;

	set ? 
      		(errstr 
			<< "The dataset "
			<< set->name()
			<< " does not have one or more of the following attributes: " << endl)
      		: (errstr
		<< " No dataset supplied. The following attributes have no values: " << endl);

	errstr
		<< " REFXCTYP REFXCRPX REFXCRVL REFXCDLT " << endl
		<< " REFYCTYP REFYCRPX REFYCRVL REFYCDLT." << endl
		<< " These are needed to convert X/Y coordinates to RA/DEC correctly." << endl
		<< " Will assume the following settings: " << endl
		<< " REFXCTYP=" << ratyp  << endl
		<< " REFXCRPX=" << refpx  << endl
		<< " REFYCRVL=" << refra  << " (reference ra)" << endl
		<< " REFXCDLT=" << radlt << endl
		<< " REFYCTYP=" << dectyp<< endl
		<< " REFYCRPX=" << refpx  << endl
		<< " REFYCRVL=" << refdec << " (reference dec)" << endl
		<< " REFYCDLT=" << decdlt << endl
		<< Sas::warning("InvalidPosWCSInfo");

	_wcs = new WCSConvertor(ratyp, refpx, refra, radlt, 
			      dectyp, refpx, refdec, decdlt);
    }
  else {
	_wcs = new WCSConvertor( set->attribute("REFXCTYP")->asString() ,
			   (unsigned) set->attribute("REFXCRPX")->asInt() ,
			   set->attribute("REFXCRVL")->asReal() ,
			   set->attribute("REFXCDLT")->asReal() ,
			   set->attribute("REFYCTYP")->asString() ,
			   (unsigned) set->attribute("REFYCRPX")->asInt() ,
			   set->attribute("REFYCRVL")->asReal() ,
			   set->attribute("REFYCDLT")->asReal() );
  }

  ENDPROC;
}

/* Destructor */
PosWCS::~PosWCS()
{
  PROC;
  errstr 
    << "Destroying PosWCS object ... "
    << Sas::message(Msg::App, Msg::Verbose);

  delete _wcs;
  ENDPROC;
}

/* Returns a pointer to WCS object */
const WCSConvertor * const PosWCS::wcs() const
{
  PROC;
//  if (!_valid)
//    {
//      errstr 
//	<< "The WCS information for POS (X/Y) coordinates being requested is not valid. This may affect any operations involving such coordinates. "
//	<< Sas::warning("InvalidPosWCSInfo");
//    }
  RETURN _wcs;
}

/* Returns the default RA */
const double PosWCS::refra() const
{
  PROC;
  RETURN _refra;
}
/* Returns the position DEC */
const double PosWCS::refdec() const
{
  PROC;
  RETURN _refdec;
}
/* Returns the position angle */
const double PosWCS::pa() const
{
  PROC;
  RETURN _refpa;
}
/* Returns whether to use OAF attitude values */
const bool PosWCS::useODF() const
{
  PROC;

  RETURN _useodf;
}

ostream & operator<< (ostream & os, const PosWCS & wcs) {
	return	os << "RA: " << wcs._refra << "DEC: " << wcs._refdec << "PA: " << wcs._refpa << endl;
}
