// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//      PsfPosition.cc:       --- Implementation of <PsfPosition> class

#include "errstr.h"
#include "param.h"
#include "PsfPosition.h"
#include "PositionUtils.h"
#include "ASCregion.h"
#include "DataSubspace.h"
#include "ExceptionErrorHandler.h"
#include <exception>

void PsfPosition::setDataSet(const string & dataSetName, const string & expr, bool init) {
	_dataSetName = dataSetName;
	_dataSet = dataSetServer->open(_dataSetName, DataSet::Read);

	if (init && _dataSet) initialise(expr);

	// need to close the dataset
	dataSetServer->close(_dataSet);
	_dataSet = 0;

} // PsfPosition::setDataSet(const string & dataSetName)

void PsfPosition::setDataSet(DataSet * pDataSet, const string & expr, bool init) {
        _dataSetName = "Unknown";
	_dataSet = pDataSet;

//        if (init && _dataSet) initialise(expr);
	if (init) initialise(expr);

} // PsfPosition::setDataSet(const string & dataSetName)

void PsfPosition::setRegion(Block *table) throw (PsfRegion::InvalidRegion) {
	GeoRegion * preg = _region;	//*preg is <GeoRegion> an object that match the type in InASCregionTable::operator>> and GeoRegion::operator=

        if (table) {
                DataSubspace * dss = new DataSubspace(table);
		bool global = false;
                for (unsigned i = 0; i < dss->components() && !global; i++) {
                        DScomp * dscomp = dss->component(i);
                        for (unsigned j = 0; j < dscomp->filters(); j++) {
                                DSfilter * dsfilter = dscomp->filter(j);
				if (dsfilter->type() != DSS::RegionBlockT && dsfilter->type() != DSS::RegionT) continue;

				global = dss->isGlobal(*dsfilter);
                                if (dsfilter->type() == DSS::RegionBlockT) {
                                        InASCregionTable ascTab(dsfilter->blockSpec());
                                        ascTab >> *preg;
                                } else {
                                        *preg = dsfilter->region();
                                }

                                DSS::DSfilterComp axes = dsfilter->comps();
                                // set the coordinate type
                                if ( axes[0].find("DET") != string::npos )           // DETX/DETY 0.05 coordinate
                                        _region->setCoordType("DET");
                                else if ( axes[0].find("RAW") != string::npos )      // RAWX/RAWY coordinate
                                        _region->setCoordType("RAW");
                                else 
					_region->setCoordType("POS");                                // X/Y coordinate

				_region->setExpression();
                                _region->setBox();

				_expr = _region->getExpression();
                        }
                }
        }

	if (_region->isEmpty()) throw PsfRegion::InvalidRegion();
} // PsfPosition::setRegion()

void PsfPosition::setPosition() {
	if (_region->isEmpty()) return;

	/* Pointing */
        double rapnt;
        double decpnt;
        double papnt;
	bool useodf;

	if (_dataSet) {
	        rapnt = _dataSet->attribute("RA_PNT")->asReal();
	        decpnt = _dataSet->attribute("DEC_PNT")->asReal();
	        papnt = _dataSet->attribute("PA_PNT")->asReal();
		useodf = booleanParameter("useodf");

	        /* Observation */
        	_datamode = _dataSet->attribute("DATAMODE")->asString();
 	} else {
		rapnt = 0;
		decpnt = 0;
		papnt = realParameter("rotate");
		if (papnt<0 && _region->getCoordType() == "POS") {
			papnt = 0;
			errstr << "PA angle is not set ('rotate') with XY coordinate system. Assume to 0" << endl
				<< Sas::warning("InvalidPA");
		}
		useodf = false;
		_datamode = "IMAGING";
	}

	_posWCS = new PosWCS(_dataSet, rapnt, decpnt, papnt, useodf);
	PositionUtils posUtils(_instrumentID);
	SourcePosition srcPos(0, 0, 0, 0);
	srcPos = posUtils.sourcePosition(_region->getCoordType(), _datamode, _obsStart, _region->getCoverBox()->refPoint().x(), _region->getCoverBox()->refPoint().y(), _posWCS, _ccdnr);

	_theta = srcPos.theta();
	_phi = srcPos.phi();
	_detx = srcPos.xdet();
	_dety = srcPos.ydet();
} // PsfPosition::setPosition()

void PsfPosition::readWindowMode() 
{
	string submode = " ";
     
	// Read the observing mode from the SUBMODE keyword
	if (_dataSet->hasAttribute("SUBMODE"))
        	submode = _dataSet->attribute("SUBMODE")->asString();
      	else
        	submode = " ";

     	// Read the data collection mode from the DATAMODE keyword 
      	string datamode = " ";
      	if (_dataSet->hasAttribute("DATAMODE"))
        	datamode = _dataSet->attribute("DATAMODE")->asString();

     	// If the observing mode is timing and the datamode is imaging
        // then the spectrum has been taken from the outer chips of
        // a timing mode obs which are actually in imagind mode 
      	if (datamode=="IMAGING" &&
           (submode=="FastUncompressed" || submode=="FastCompressed") )
               submode= "PrimeFullWindow";

     	// Check if submode string is ok and if not set to default
      	ExceptionErrorHandler local;
      	ErrorHandler* oldHandler = errHandler.handler(&local);
      	try {
            	_windowMode = Xmm::toMode(submode);
      	} catch(SasException::Error) { 
		errstr<<"SUBMODE not valid or missing "<<
          	"- using FullWindow" <<Sas::warning("UnknownModeString");
        	submode = "PrimeFullWindow";
        	_windowMode = Xmm::toMode(submode); // d/f
      	}
      	errHandler.handler(oldHandler);
} // PsfPosition::readWindowMode()

void PsfPosition::setCcdnr() {
	int ccdnr = intParameter("ccdnr");

        if (ccdnr)
                _ccdnr = ccdnr;
        else if (_dataSet && _dataSet->hasAttribute("CCDNR")) {
                _ccdnr = _dataSet->attribute("CCDNR")->asInt();
		ccdnr = _ccdnr;
	}
        else if (_dataSet) {
		
        // If the CCD number isn't given set it to the central CCD for the
        //  instrument. This is necessary for timing mode
                DataSubspace * dss = new DataSubspace(_dataSet->block(0));
		if (dss->hasFilter("CCDNR", DSS::RangeT)) {
                        DSfilter * dscomp = dss->component(0)->filter("CCDNR", DSS::RangeT);
				
			if (dscomp != 0) {
				Range<double> r = dscomp->ranges().front();
        	               	double up, low;
                	       	RangeBound upB, lowB;
                       		r.values(up, low, upB, lowB);
				_ccdnr = static_cast<int>(up);
			}
			ccdnr = _ccdnr;
		} else if (instrumentID() == Xmm::EPN)  //default EPN center ccd
                        _ccdnr = 4;
                else _ccdnr = 1;  //default MOS center ccd
        }
	else if (instrumentID() == Xmm::EPN)  //default EPN center ccd
                        _ccdnr = 4;
        else _ccdnr = 1;  //default MOS center ccd

	if (!ccdnr && _region->getCoordType() == "RAW") 
		errstr << "No CCD number supplied ('ccdnr') with RAW coordinate system. Assume to " << _ccdnr << endl
			<< Sas::warning("InvalidCCDNR");
			

} // PsfPosition::setCcdnr()

void PsfPosition::initialise( const string & expr) {
	_expr = expr;

	// Dates
	if (_dataSet) {
	      	_obsStart = _dataSet->attribute("DATE-OBS")->asString();

      		if (_dataSet->hasAttribute("DATE-END")) 
			_obsEnd = _dataSet->attribute("DATE-END")->asString();
		else {
			// No end value so set it to be 10 ks after observation start
			STime numdate(_obsStart);
			numdate = numdate + 10000.0;
			_obsEnd = numdate.toString();
		}
	
		// Get instrument ID
		_instrument = _dataSet->attribute("INSTRUME")->asString();
	} else {
		string inst = stringParameter("instrument");
		
		_instrument = inst.empty() ? "M1" : inst;
		_obsStart = "1999-12-10T14:32:00.000";
	} 
	_instrumentID = Xmm::toInstrument(_instrument);
	calServer.setInstrument(_instrumentID);

     	// set region information
	if (_region) delete _region;
	_region  = new PsfRegion();

        string rgnStr = stringParameter("region");

	try {
		if (!(_expr.empty() && rgnStr.empty())) {	// region information supplied by user
			if (_expr.empty()) _expr = rgnStr;
			_region->setRegion(_expr);
		} else if (_dataSet) {	//region information given in the dataset
			Block * table = _dataSet->block(0);
			setRegion(table);
		} else {	// no region no dataset
			_region->setRegion("(DETX, DETY) IN circle(0, 0, 2000)");
		}
	} catch (PsfRegion::InvalidRegion) {
			errstr << "Invalid REGION: '" << _region->getExpression() << "'" << endl
				<<" Set to default (DETX, DETY) IN circle(0, 0, 2000)" << endl
				<< Sas::warning("InvalidRegionInfo");
			_region->setRegion("(DETX, DETY) IN circle(0, 0, 2000)");
	}
	
        // Get the CCD number
	setCcdnr();

	setPosition();

	// get window mode
	if (_dataSet) readWindowMode();	
} //PsfPosition::initialise()

double PsfPosition::maxDist(double v1, double v2, double v3, double v4) const {
	double max = v1, min = v2;
	if (max < min) max = v2, min = v1;
	if (v3 > max) max = v3;
	else if (v3 < min) min = v3;
	if (v4 > max) max = v4;
	else if (v4 < min) min = v4;

	return (max-min);
}

CoverFBox PsfPosition::coverBox() const {
	// this function return a box that covers the region. the center is in TELCOORD. The unit of theta and phi are arcsecs
	// and degree respectively. The size is in DET005.
	if (_region->isEmpty()) return CoverFBox();

	const string type = _region->getCoordType();
        const FBox * box = _region->getCoverBox();
	double xOrgSize = box->xWidth(), yOrgSize = box->yWidth();
	double xSize = xOrgSize / 2.0, ySize = yOrgSize / 2.0;
	double pa = 0.0;
	CoordinateType coorType = CoorType::UNDEFINE;
	FPoint center = box->refPoint();
	CoordPair leftBottom(center.x() - xSize, center.y() - ySize), rightTop(center.x() + xSize, center.y() + ySize);
	CoordPair leftTop(center.x() - xSize, center.y() + ySize), rightBottom(center.x() + xSize, center.y() - ySize);
//cout << xSize << " " << ySize << endl; 
        PositionUtils posUtils(_instrumentID);
	
	if (type == "POS") {
		SourcePosition lb = posUtils.sourcePosition("POS", _datamode, _obsStart, leftBottom.x(), leftBottom.y(), _posWCS, _ccdnr);
		SourcePosition rt = posUtils.sourcePosition("POS", _datamode, _obsStart, rightTop.x(), rightTop.y(), _posWCS, _ccdnr);
                SourcePosition lt = posUtils.sourcePosition("POS", _datamode, _obsStart, leftTop.x(), leftTop.y(), _posWCS, _ccdnr);
                SourcePosition rb = posUtils.sourcePosition("POS", _datamode, _obsStart, rightBottom.x(), rightBottom.y(), _posWCS, _ccdnr);
//printf("lb(%f, %f) rt(%f, %f) lt(%f, %f) rb(%f, %f)\n", lb.xdet(), lb.ydet(), rt.xdet(), rt.ydet(), lt.xdet(), lt.ydet(), rb.xdet(), rb.ydet()); 
//cout << xSize << " " << ySize << endl;
//		xSize = maxDist(lb.xdet(), rt.xdet(), lt.xdet(), rb.xdet());
//		ySize = maxDist(lb.ydet(), rt.ydet(), lt.ydet(), rb.ydet());
		xOrgSize = sqrt(pow(lb.xdet() - rb.xdet(), 2) + pow(lb.ydet() - rb.ydet(), 2));
		yOrgSize = sqrt(pow(lb.xdet() - lt.xdet(), 2) + pow(lb.ydet() - lt.ydet(), 2));
		xSize = ySize = sqrt(pow(lb.xdet() - rt.xdet(), 2) + pow(lb.ydet() - rt.ydet(), 2));
		
//cout << xSize << " " << ySize << endl;
//		xRotAngle = acos((rb.xdet() - lb.xdet()) / box->xWidth());
//		yRotAngle = asin((lb.ydet() - rb.ydet()) / box->xWidth());
//		xRotAngle = 1 - 2 * (lb.theta() * cos(lb.phi()) > rb.theta() * cos(rb.phi()));
//		yRotAngle = 1 - 2 * (lb.theta() * sin(lb.phi()) > lt.theta() * sin(lt.phi()));
//                string sLevel(stringParameter("level"));

//                if (sLevel.compare("ELLBETA")) xRotAngle = -1.0;
//		xRotAngle = -1.0;	// version 1.1.4 or older
		coorType = CoorType::XY;

		pa = _posWCS->pa();
//cout << xRotAngle / M_PI * 180 << " " << yRotAngle / M_PI * 180 << endl;
//		xSize *= 2.0;
//     		ySize *= 2.0;
	} else if (type == "RAW") {
                CoordPair lb = posUtils.RawTodetxy005(leftBottom.x(), leftBottom.y(), _ccdnr);
                CoordPair rt = posUtils.RawTodetxy005(rightTop.x(), rightTop.y(), _ccdnr);
 	        CoordPair lt = posUtils.RawTodetxy005(leftTop.x(), leftTop.y(), _ccdnr);
		CoordPair rb = posUtils.RawTodetxy005(rightBottom.x(), rightBottom.y(), _ccdnr);

		coorType = CoorType::RAW;
//		xSize = maxDist(lb.x(), rt.x(), lt.x(), rb.x());
//		ySize = maxDist(lb.y(), rt.y(), lt.y(), rb.y());
		
		xOrgSize = (xOrgSize - 0.5) * 22., yOrgSize = (yOrgSize - 0.5) * 22.;
		// for PN, the pixel size is 4.1" rather than 1.1" as the EMOS camera and TELCOORD/CAMCOORD
		// special treatment for PN RAW data
		if (_instrumentID == Xmm::EPN)
			xOrgSize *= 4.1 / 1.1, yOrgSize *= 4.1 / 1.1;
                xSize = ySize = sqrt(pow(xOrgSize, 2) + pow(yOrgSize, 2));

//		xRotAngle = acos((rb.x() - lb.x()) / box->xWidth());
//		yRotAngle = asin((lb.y() - rb.y()) / box->xWidth());
//cout << rt.x() << " " << rt.y() << " -- " << lb.x() << " " << lb.y() << endl;
//		xRotAngle = 1 - 2 * (rt.x() > lb.x());
//		yRotAngle = 1 - 2 * (rt.y() > lb.y());

//		string sLevel(stringParameter("level"));

//		if (!sLevel.compare("ELLBETA")) xRotAngle = -1.0;
		
		switch (_instrumentID) {
		case Xmm::EPN	:	if (_ccdnr >= 9 && _ccdnr <= 12) pa = 90.0;
					else if (_ccdnr > 0) pa = 270.0;
					else {
						cout << "Not supported ccd number. Set rotation disabled." << endl; pa = -0.5;
					}
					break;
		case Xmm::EMOS1	:	if (_ccdnr == 1) pa = 0.0;
					else if (_ccdnr >=5 && _ccdnr <=7) pa = 270.0;
					else if (_ccdnr >=2 && _ccdnr <=4) pa = 90.0;
                                        else {
                                                cout << "Not supported ccd number. Set rotation disabled." << endl; pa = -0.5;
                                        }
                                        break;
		case Xmm::EMOS2	:	if (_ccdnr == 1) pa = 270.0;
                                        else if (_ccdnr >=5 && _ccdnr <=7) pa = 180.0;
                                        else if (_ccdnr >=2 && _ccdnr <=4) pa = 0.0;
                                        else {
                                                cout << "Not supported ccd number. Set rotation disabled." << endl; pa = -0.5;
                                        }
                                        break;
                default         :       cout << "Not supported instrument. Set rotation disabled." << endl; pa = -0.5;
		}
	} else if (type == "DET") {
		ySize = xSize = sqrt(pow(xOrgSize, 2) + pow(yOrgSize, 2));
		coorType = CoorType::DET;
//                xRotAngle = 1 - 2 * (rightTop.x() > leftTop.x());
//                yRotAngle = 1 - 2 * (rightTop.y() > rightBottom.y());

//                string sLevel(stringParameter("level"));

//                if (!sLevel.compare("ELLBETA")) xRotAngle = -1.0;
                
		switch (_instrumentID) {
		case Xmm::EPN 	: 	pa = 90.0; break;
		case Xmm::EMOS1	:	pa = 0.0; break;
		case Xmm::EMOS2	:	pa = 270.0; break;
		default		:	cout << "Not supported instrument. Set rotation disabled." << endl; pa = -0.5;
		}
//		ySize *= 2.0;
//		xSize *= 2.0;
	}

	// if cos(xRotAngle) < 0, need to flipped along Y axis. if sin(yRotAngle) < 0, need to flipped along X axis.
	// after flipping, rotate clockwise RotAngle, where RotAngle=acos(abs(cos(xRotAngle))).
        // actually in mathmatics, the coordinate conversion does not need to think of the flippings, just use the matrix multiplicaiton.
//	if (yRotAngle < 0) yRotAngle += (2 * M_PI);

//	return CoverFBox(FPoint(_theta, _phi * 180 / M_PI), xSize, ySize, xOrgSize, yOrgSize, pa, acos(xRotAngle) * 180 / M_PI, acos(yRotAngle) * 180 / M_PI);	// version 1.1.4 or older
	return CoverFBox(FPoint(_theta, _phi * 180 / M_PI), xSize, ySize, xOrgSize, yOrgSize, pa, coorType);
//	return CoverFBox(FPoint(_theta, _phi * 180 / M_PI), xSize, ySize, xRotAngle, yRotAngle);
}

ostream & operator<< (ostream & out, const CoverFBox & box) {
	out << "Original size in DET005: (" << box.xOrgDetSize() << ", " << box.yOrgDetSize() << ")" << endl;
	out << "Cover Box: center -- (" << box.refPoint().x() << ", " << box.refPoint().y() << ")    size -- (" << box.xWidth() << ", " <<  box.yWidth() << ")" << endl;
	out << "Rotate: " << box.pa() << " Coordinate Type: " << CoorType::toString(box.coor()) << endl;
/*	switch (box.coorType()) {
		case CoorType::XY:	out << "XY SKY"; break;
		case CoorType::DET:	out << "DETX/Y"; break;
		case CoorType::RAW:	out << "RAWX/Y"; break;
		default:		out << "UNDEFINED COORDINATE";
	}
	out << endl;
*/
	return out;
}

ostream & operator<< (ostream & out, const PsfPosition & pos) {
	out << "Expression: " << pos.expression() << endl;
	out << "from Image Set: " << pos.dataSet() << endl;
	out << "Instrument: " << pos.instrument() << endl;
	out << "CCDNR: " << pos.ccdnr() << endl;
	out << "Time: " << pos.obsStart() << " to " << pos.obsEnd() << endl;
	const PsfRegion * reg = pos.region();
	out << "PSF center: (" << pos.theta() << ", " << pos.phi() << ") in TELCOORD, (" << pos.detx() << ", " << pos.dety() << ") in DET005" << endl;
	out << *reg;

	return out;
}

ostream & operator<< (ostream & out, const CoorType::Type & coor) {
	return out << CoorType::toString(coor);
}
